#
#   Japanese charcter category map
#
#   $Id: char.def 9 2012-12-12 04:13:15Z togiso $;
#

###################################################################################
#
#  CHARACTER CATEGORY DEFINITION
#
#  CATEGORY_NAME INVOKE GROUP LENGTH
#
#   - CATEGORY_NAME: Name of category. you have to define DEFAULT class.
#   - INVOKE: 1/0:   always invoke unknown word processing, evan when the word can be found in the lexicon
#   - GROUP:  1/0:   make a new word by grouping the same chracter category
#   - LENGTH: n:     1 to n length new words are added
#
DEFAULT         0 1 0  # DEFAULT is a mandatory category!
SPACE           0 1 0
KANJI           0 0 2
SYMBOL          1 1 0
NUMERIC         1 1 0
ALPHA           1 1 0
HIRAGANA        0 1 2
KATAKANA        1 1 2
KANJINUMERIC    0 1 0  #change INVOKE 1->0
GREEK           1 1 0
CYRILLIC        1 1 0

###################################################################################
#
# CODE(UCS2) TO CATEGORY MAPPING
#

# SPACE
0x0020 SPACE  # DO NOT REMOVE THIS LINE, 0x0020 is reserved for SPACE

# ASCII
0x0030..0x0039 NUMERIC  #0-9

# KANJI-NUMERIC (〇 一 二 三 四 五 六 七 八 九 十 百 千 万 億 兆)
0x3007 KANJINUMERIC KANJI
0x4E00 KANJINUMERIC KANJI
0x4E8C KANJINUMERIC KANJI
0x4E09 KANJINUMERIC KANJI
0x56DB KANJINUMERIC KANJI
0x4E94 KANJINUMERIC KANJI
0x516D KANJINUMERIC KANJI
0x4E03 KANJINUMERIC KANJI
0x516B KANJINUMERIC KANJI
0x4E5D KANJINUMERIC KANJI
0x5341 KANJINUMERIC KANJI
0x767E KANJINUMERIC KANJI
0x5343 KANJINUMERIC KANJI
0x4E07 KANJINUMERIC KANJI
0x5104 KANJINUMERIC KANJI
0x5146 KANJINUMERIC KANJI
